@@ -130,6 +130,21 @@ def _dl(src: str, dest: str, mode: str, show_: bool, logger: logging.Logger, ret
     remote, s_path = _parse_remote_path(src)
 
     kubectl = Executable('kubectl')
+
+    # Option `--pod-running-timeout` in kubectl run/exec doesn't help - exec errors out with:
+    #  error: Internal error occurred: unable to upgrade connection: container not found ("container-name")
+    # Apparently pod gets Running status without waiting for a running container. So must wait explicitly
+    # Unfortunately kubectl wait command doesn't accept the container option, adding another wrinkle to the opt parse
+    waitable = remote
+    options = ['--container', '-c']
+    if container_option := next((x for x in remote.split() if x in options), None):
+        i = remote.index(container_option)
+        waitable = remote[:i] + remote[i:].split(maxsplit=2)[-1]
+    elif container_option := next((x for x in waitable.split() if any((y for y in options if x.startswith(y)))), None):
+        i = remote.index(container_option)
+        waitable = remote[:i] + remote[i:].split(maxsplit=1)[-1]
+    kubectl.stream('wait --for=condition=ready pod --timeout=120s', waitable)
+
     resp = kubectl.run('exec', remote, '-- du -sh', s_path, check=False)
     if resp.returncode:
         logger.error(resp.stderr)
@@ -328,7 +343,11 @@ def ensure_namespace(mode: str, logger: logging.Logger, *, namespace: Optional[s
     raise PermissionError(namespace_exists.stderr)
 
 
-def ensure_pvc(spec: dict, logger: logging.Logger) -> None:
+def ensure_pvc(
+        spec: dict,
+        logger: logging.Logger,
+        _retries: int = 1  # noqa internal undocumented parameter for retrying PVC creation
+) -> None:
     """Ensure PVC exists in k8s, create if not present
     :param spec: dict with PVC parameters
     :param logger: logger object
@@ -345,9 +364,15 @@ def ensure_pvc(spec: dict, logger: logging.Logger) -> None:
         from .templates import persistent_volume_claim
         _, path = persistent_volume_claim.generate_template(spec)
         ensure_namespace(spec['MODE'], logger, namespace=spec['NAMESPACE'])
-        kubectl.stream('create --filename', path)
+        import subprocess
+        try:
+            kubectl.stream('create --filename', path)
+        except subprocess.CalledProcessError as e:
+            if _retries and 'terminated' in e.stderr:
+                log.get_logger(name='plain').error(e.stderr)
+                ensure_pvc(spec, logger, 0)
 
-    pvc_desired_states = [x.lower() for x in spec.get('PERSISTENT_VOLUME_CLAIM_DESIRED_STATES') or ['Bound']]
+    pvc_desired_states = [x for x in spec.get('PERSISTENT_VOLUME_CLAIM_DESIRED_STATES') or ['Bound']]
     pvc_name = spec['PERSISTENT_VOLUME_CLAIM_NAME']
 
     pvc_phase_jsonpath = 'jsonpath="{.status.phase}"'
@@ -436,7 +461,6 @@ def create_oneliner_job(
     :return: job name
     """
     from . import strings
-    from .templates import batch_job
 
     logger = log.get_logger(name=spec.get('LOGGER_NAME'), level=spec.get('LOG_LEVEL') or 'INFO')
     ensure_pvc(spec, logger)
@@ -459,7 +483,16 @@ def create_oneliner_job(
         spec['JOB_NAME'] = trunc(spec['JOB_NAME'] + f'-{suf}')
 
     manifests_folder = spec.setdefault('GENERATED_MANIFESTS_FOLDER', spec['CACHE_FOLDER'])
+
+    from .templates import batch_job, recursive_has_pair
+
     spec['JOB_NAME'], path = batch_job.generate_template(spec, manifests_folder, filename=container_name)
+    if wait_offset := spec.get('WAIT_BEFORE_IMAGE_PULL_POLICY_ALWAYS', 0.1):
+        if recursive_has_pair(spec, 'IMAGE_PULL_POLICY', 'Always'):
+            from random import random
+            sleep = wait_offset + random() * 10  # nosec CWE-330
+            logger.info(f'pullImagePolicy=Always detected!\n{sleep=:.2f}s to avoid thundering herd DDoS')
+            time.sleep(sleep)
     kubectl.stream('create --filename', path)
     log.redact_file(path, redact)
 
@@ -487,8 +520,6 @@ def await_k8s_job_completion(spec: dict, tail: Optional[int] = None) -> bool:
     logger = log.get_logger(name=spec.get('LOGGER_NAME'))
     logger_plain = log.get_logger('plain')
     job_status_cmd = 'get job --output jsonpath={.status} ' + name
-    output_wide_on_debug = '--output wide' if spec.get('LOG_LEVEL') == 'DEBUG' else ''
-    get_pods_cmd = f'get pods --ignore-not-found --selector=job-name={name} {output_wide_on_debug}'
 
     kubectl = Executable('kubectl', f'kubectl --namespace {namespace}')
     kubectl_run = functools.partial(kubectl.run, show_cmd=False, check=False)
@@ -506,6 +537,14 @@ def await_k8s_job_completion(spec: dict, tail: Optional[int] = None) -> bool:
         logger_plain.warning(f'{response.stderr}\nWaiting {backoff}s')
         response = kubectl_run(job_status_cmd)
 
+    get_pods_cmd = f'get pods --ignore-not-found --selector=job-name={name}'
+    kubectl.stream(
+        get_pods_cmd,
+        r"""--output jsonpath='{range .items[*].status.containerStatuses[*]}{.image} -> {.imageID}{"\n"}{end}'""",
+        show_cmd=False,
+    )
+
+    get_pods_cmd += ' --output wide' if spec.get('LOG_LEVEL') == 'DEBUG' else ''
     kubectl.stream(get_pods_cmd)
 
     import itertools
@@ -758,3 +797,102 @@ def is_path_local_best_effort(path: str | Path) -> bool:
         return False
 
     return True
+
+
+def fetch_from_image(namespace: str, image: str, source, target: str, mode: str) -> None:
+    logger = log.get_logger()
+    ensure_namespace(mode, logger, namespace=namespace)
+    kubectl = Executable('kubectl', 'kubectl --namespace', namespace)
+
+    from random import getrandbits
+    pod_name = Path(target).name.split('.', maxsplit=1)[0] + '-temp-' + str(getrandbits(16)) # nosec: B311
+
+    kubectl.stream('delete pod --ignore-not-found --wait', pod_name, mode=mode)
+    kubectl.stream('run --image-pull-policy=Always --image', image, pod_name, '--command -- sleep 3600', mode=mode)
+    if mode == 'dry-run':
+        return
+    kubectl.stream('get pods', pod_name, '--output',
+                   r"""jsonpath='{range .status.containerStatuses[*]}{.image} -> {.imageID}{"\n"}{end}'""")
+    kubectl_cp(f'--{namespace=} {pod_name}:{source}', target, mode=mode)
+    kubectl.stream('delete pod --ignore-not-found --wait=false', pod_name, mode=mode)
+
+
+def prep_binary(mode: str, spec: dict, name: str, refresh_rate_default) -> str:
+    """Locate a usable binary, fetch from image if missing. Order of precedence:
+
+    1. name in manifest as {name.upper()}_PATH or {name}Path
+    2. name in PATH
+    3. name in temp_dir
+    4. name in provided k8s image
+
+    :param mode: execution mode
+    :param spec: spec dict
+    :param name: name of the binary
+    :param refresh_rate_default: the rate at which binary will be refreshed from k8s image pull
+    :raises: ValueError if path not found
+    :raises: PermissionError if path not executable
+    """
+    import shutil
+    path = spec.get(f'{name}_PATH'.upper()) or spec.get(f'{name}Path') or shutil.which(name)
+
+    if not path:
+        temp_dir = spec['CACHE_FOLDER']
+        image = spec['JOB_IMAGE']
+        namespace = spec['NAMESPACE']
+        path_on_image = spec['PATH_ON_IMAGE']
+        image_pull_policy = spec.get('IMAGE_PULL_POLICY')
+
+        logger = log.get_logger()
+        logger.warning(f'{name} path not set in manifest or found in PATH. Checking in temp dir')
+        path = os.path.join(temp_dir, name)
+
+        set_image_pull_policy_default(spec, refresh_rate_default)
+        if image_pull_policy == 'Always':
+            logger.warning(f'Force updating from k8s image')
+            swap_path = path + '.swap'
+            fetch_from_image(namespace, image, path_on_image, swap_path, mode)
+            if mode == 'dry-run':
+                return path
+            shutil.move(swap_path, path)
+        if not os.path.exists(path):
+            logger.warning(f'{name} path not found in temp dir, fetching from k8s image')
+            fetch_from_image(namespace, image, path_on_image, path, mode)
+            if mode == 'dry-run':
+                return path
+    if not path:
+        raise ValueError(f'{name} executable not found')
+    if not os.path.isabs(path):
+        path = os.path.abspath(path)
+    if not os.access(path, os.X_OK):
+        import subprocess
+        subprocess.run(f'chmod +x {path}', shell=True)
+    if not os.access(path, os.X_OK):
+        raise PermissionError(f'{name} path is not executable: {path}')
+    return path
+
+
+# Edge case, which looks like a bug:
+#  1. node1 runs with "Always" and updates version. Next run:
+#  2. node2 runs with "IfNotPresent" and doesn't update
+#  This results in a later run will surprisingly use an older version.
+#  Assuming the older version is still good, the run will be successful.
+#  It's not ideal but implementing a solution here, we can always ensure version is latest by manually setting "Always"
+def set_image_pull_policy_default(spec: dict, refresh_rate_default: float):
+    from random import random
+    from basepak.templates import recursive_has_pair
+
+    if random() * 99.99 > spec.get('REFRESH_RATE', refresh_rate_default):  # nosec: B311
+        return
+    spec.setdefault('IMAGE_PULL_POLICY', 'Always')
+
+    wait_offset = spec.get('WAIT_BEFORE_IMAGE_PULL_POLICY_ALWAYS', 0.1)
+    if not (wait_offset and recursive_has_pair(spec, 'IMAGE_PULL_POLICY', 'Always')):
+        return
+    sleep = wait_offset + random() * 10  # nosec CWE-330
+    logger = log.get_logger()
+    logger.info(f'pullImagePolicy=Always detected!\n{sleep=:.2f}s to avoid thundering herd DDoS')
+    if spec.get('MODE', '') == 'dry-run':
+        logger.info('Dry run. Skipping wait..')
+        return
+
+    time.sleep(sleep)
