@@ -29,7 +29,6 @@ POD_SPEC_DEFAULT = {
             ]}}
 }
 
-
 def generate_template(
         params: Mapping, dump_folder: Optional[str | Path] = None, filename: Optional[str] = None
 ) -> tuple[str, str]:
@@ -81,8 +80,7 @@ def generate_template(
                 'spec': {
                     'containers': [{
                         'name': params.get('CONTAINER_NAME') or job_name,
-                        # stable instead of latest, to avoid unnecessary pulls
-                        'image': params.get('JOB_IMAGE') or params.get('DEFAULT_IMAGE') or 'busybox:stable',
+                        'image': params.get('JOB_IMAGE') or params.get('DEFAULT_IMAGE') or consts.DEFAULT_IMAGE,
                         'volumeMounts': [{
                                 'name': volume_name,
                                 'mountPath': params['JOB_MOUNT_PATH'],
@@ -90,6 +88,7 @@ def generate_template(
                         'command': params.get('COMMAND') or [],
                         'args': params.get('ARGS') or [],
                         'env': params.get('ENV_VARS') or [],
+                        **({'imagePullPolicy': params['IMAGE_PULL_POLICY']} if params.get('IMAGE_PULL_POLICY') else {}),
                         **security_context,
                     }],
                     **pod_spec,
