@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import logging
 import subprocess
+import os
 from contextlib import contextmanager
 from unittest.mock import MagicMock, patch
 
@@ -10,6 +11,7 @@ import pytest  # type: ignore
 from basepak import k8s_utils
 from basepak.time import sleep
 from basepak.versioning import Version
+from basepak import consts
 
 SUPPORTED_MODES = ['dry-run', 'normal', 'unsafe']
 
@@ -267,11 +269,13 @@ def test_kubectl_upload_download_file(tmp_path, mode):
         tmp_file = tmp_path / 'tmp.yaml'
         tmp_file.write_text('test content')
 
-        remote_path = f'{pod}:/tmp/file'
+        remote_path = f'{pod}:/tmp/file --container {pod} '
+        # upload
         k8s_utils.kubectl_cp(src=tmp_file, dest=remote_path, mode='unsafe' if mode == 'dry-run' else mode, retries=1)
 
         tmp_file.unlink()
 
+        # download
         k8s_utils.kubectl_cp(dest=tmp_file, src=remote_path, mode=mode, retries=1)
 
 @pytest.mark.parametrize(
@@ -290,7 +294,7 @@ def test_get_kubectl_version():
 @pytest.mark.parametrize('mode', SUPPORTED_MODES)
 def test_kubectl_upload_download_dir(tmp_path, mode):
     with _fresh_pod() as pod:
-        remote_path = f'{pod}:/tmp/dir'
+        remote_path = f'-c{pod} {pod}:/tmp/dir'
         local_dir = tmp_path / 'dir'
         local_dir.mkdir()
         for i in range(100):
@@ -347,3 +351,23 @@ def test_await_k8s_job_completion_complete_with_fail():
         'MODE': 'unsafe',
     })
     subprocess.run(f'kubectl delete job {name} --wait=false', shell=True, check=False)
+
+@pytest.mark.parametrize('mode', SUPPORTED_MODES)
+@pytest.mark.parametrize('refresh_rate', (0,10,100))
+def test_prep_binary(mode, tmp_path, refresh_rate):
+    name = 'shred'
+    spec = {
+        'CACHE_FOLDER': tmp_path,
+        'NAMESPACE': 'default',
+        'JOB_IMAGE': consts.DEFAULT_IMAGE,
+        'PATH_ON_IMAGE': f'/bin/{name}',
+    }
+    k8s_utils.prep_binary(mode=mode, spec=spec, name=name, refresh_rate_default=refresh_rate)
+
+    if mode != 'dry-run':
+        assert os.access(f'{tmp_path}/{name}', os.X_OK)
+
+
+
+
+# def fetch_from_image(namespace: str, image: str, source, target: str, mode: str) -> None:
